@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/python
 
 ## Copyright (C) 2007, 2008, 2009, 2010 Red Hat, Inc.
 ## Copyright (C) 2008 Novell, Inc.
@@ -31,9 +31,6 @@ import asyncipp
 from debug import *
 import debug
 
-from dbus.mainloop.glib import DBusGMainLoop
-DBusGMainLoop (set_as_default=True)
-
 CUPS_PK_NAME  = 'org.opensuse.CupsPkHelper.Mechanism'
 CUPS_PK_PATH  = '/'
 CUPS_PK_IFACE = 'org.opensuse.CupsPkHelper.Mechanism'
@@ -47,6 +44,8 @@ CUPS_PK_NEED_AUTH = 'org.opensuse.CupsPkHelper.Mechanism.NotPrivileged'
 ###### thread.
 ######
 
+_DevicesGet_uses_new_api = None
+
 ###
 ### A class to handle an asynchronous method call.
 ###
@@ -103,7 +102,9 @@ class _PK1AsyncMethodCall:
             return
 
         if str (error) == '':
+            gtk.gdk.threads_enter ()
             self._client_reply_handler (self._conn, self._unpack_fn (*args))
+            gtk.gdk.threads_leave ()
             self._destroy ()
             return
 
@@ -116,7 +117,9 @@ class _PK1AsyncMethodCall:
 
         if exc.get_dbus_name () == CUPS_PK_NEED_AUTH:
             exc = cups.IPPError (cups.IPP_NOT_AUTHORIZED, 'pkcancel')
+            gtk.gdk.threads_enter ()
             self._client_error_handler (self._conn, exc)
+            gtk.gdk.threads_leave ()
             self._destroy ()
             return
 
@@ -151,7 +154,10 @@ class _WriteToTmpFile:
     def __init__ (self, kwds, reply_handler, error_handler):
         self._reply_handler = reply_handler
         self._error_handler = error_handler
-        (tmpfd, tmpfname) = tempfile.mkstemp ()
+
+        # Create the temporary file in /tmp to ensure that
+        # cups-pk-helper-mechanism is able to write to it.
+        (tmpfd, tmpfname) = tempfile.mkstemp (dir="/tmp")
         os.close (tmpfd)
         self._filename = tmpfname
         debugprint ("Created tempfile %s" % tmpfname)
@@ -209,8 +215,8 @@ class PK1Connection:
             # No system D-Bus.
             self._system_bus = None
 
-        self._devicesget_uses_new_api = None
-        if self._system_bus:
+        global _DevicesGet_uses_new_api
+        if _DevicesGet_uses_new_api == None and self._system_bus:
             try:
                 obj = self._system_bus.get_object(CUPS_PK_NAME, CUPS_PK_PATH)
                 proxy = dbus.Interface (obj, dbus.INTROSPECTABLE_IFACE)
@@ -232,7 +238,7 @@ class PK1Connection:
 
                             num_args += 1
 
-                        self._devicesget_uses_new_api = num_args == 4
+                        _DevicesGet_uses_new_api = num_args == 4
                         debugprint ("DevicesGet new API: %s" % (num_args == 4))
                         break
 
@@ -368,7 +374,8 @@ class PK1Connection:
         return None
 
     def getDevices (self, *args, **kwds):
-        if self._devicesget_uses_new_api:
+        global _DevicesGet_uses_new_api
+        if _DevicesGet_uses_new_api:
             (use_pycups, reply_handler, error_handler,
              tup) = self._args_kwds_to_tuple ([int, int, list, list],
                                               [("timeout", 0),
@@ -377,26 +384,26 @@ class PK1Connection:
                                                ("exclude_schemes", [])],
                                               args, kwds)
         else:
-                (use_pycups, reply_handler, error_handler,
-                 tup) = self._args_kwds_to_tuple ([int, list, list],
-                                                  [("limit", 0),
-                                                   ("include_schemes", []),
-                                                   ("exclude_schemes", [])],
-                                                  args, kwds)
+            (use_pycups, reply_handler, error_handler,
+             tup) = self._args_kwds_to_tuple ([int, list, list],
+                                              [("limit", 0),
+                                               ("include_schemes", []),
+                                               ("exclude_schemes", [])],
+                                              args, kwds)
 
-                if not use_pycups:
-                    # Special handling for include_schemes/exclude_schemes.
-                    # Convert from list to ","-separated string.
-                    newtup = list (tup)
-                    for paramindex in [1, 2]:
-                        if len (newtup[paramindex]) > 0:
-                            newtup[paramindex] = reduce (lambda x, y:
-                                                             x + "," + y,
-                                                         newtup[paramindex])
-                        else:
-                            newtup[paramindex] = ""
+            if not use_pycups:
+                # Special handling for include_schemes/exclude_schemes.
+                # Convert from list to ","-separated string.
+                newtup = list (tup)
+                for paramindex in [1, 2]:
+                    if len (newtup[paramindex]) > 0:
+                        newtup[paramindex] = reduce (lambda x, y:
+                                                         x + "," + y,
+                                                     newtup[paramindex])
+                    else:
+                        newtup[paramindex] = ""
 
-                    tup = tuple (newtup)
+                tup = tuple (newtup)
 
         self._call_with_pk (use_pycups,
                             'DevicesGet', tup, reply_handler, error_handler,
@@ -439,7 +446,7 @@ class PK1Connection:
 
     def cancelJob (self, *args, **kwds):
         (use_pycups, reply_handler, error_handler,
-         tup) = self._args_kwds_to_tuple ([int],
+         tup) = self._args_kwds_to_tuple ([int, bool],
                                           [(None, None),
                                            (None, False)], # purge_job
                                           args, kwds)
@@ -684,4 +691,6 @@ if __name__ == '__main__':
             print "get jobs error: %s" % repr (exc)
 
     UI ()
+    from dbus.mainloop.glib import DBusGMainLoop
+    DBusGMainLoop (set_as_default=True)
     gtk.main ()
