@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/python
 
 ## Copyright (C) 2007, 2008, 2009, 2010 Red Hat, Inc.
 ## Author: Tim Waugh <twaugh@redhat.com>
@@ -25,12 +25,8 @@ import gtk
 import os
 from errordialogs import *
 from debug import *
-
-_ = lambda x: x
+from gettext import gettext as _
 N_ = lambda x: x
-def set_gettext_function (fn):
-    global _
-    _ = fn
 
 class AuthDialog(gtk.Dialog):
     AUTH_FIELD={'username': N_("Username:"),
@@ -114,6 +110,30 @@ class AuthDialog(gtk.Dialog):
         i = self.auth_info_required.index (field)
         self.field_entry[i].grab_focus ()
 
+###
+### An auth-info cache.
+###
+class _AuthInfoCache:
+    def __init__ (self):
+        self.creds = dict() # by (host,port)
+
+    def cache_auth_info (self, data, host=None, port=None):
+        if port == None:
+            port = 631
+
+        self.creds[(host,port)] = data
+
+    def lookup_auth_info (self, host=None, port=None):
+        if port == None:
+            port = 631
+
+        try:
+            return self.creds[(host,port)]
+        except KeyError:
+            return None
+
+global_authinfocache = _AuthInfoCache ()
+
 class Connection:
     def __init__ (self, parent=None, try_as_root=True, lock=False,
                   host=None, port=None, encryption=None):
@@ -135,12 +155,21 @@ class Connection:
         self._operation_stack = []
         self._lock = lock
         self._gui_event = threading.Event ()
+
+        creds = global_authinfocache.lookup_auth_info (host=host, port=port)
+        if creds != None:
+            if (creds[0] != 'root' or try_as_root):
+                (self._use_user, self._use_password) = creds
+            del creds
+
         self._connect ()
 
     def _begin_operation (self, operation):
+        debugprint ("%s: Operation += %s" % (self, repr (operation)))
         self._operation_stack.append (operation)
 
     def _end_operation (self):
+        debugprint ("%s: Operation ended" % self)
         self._operation_stack.pop ()
 
     def _get_prompt_allowed (self, ):
@@ -237,8 +266,7 @@ class Connection:
 
                     raise
             except cups.HTTPError, (s,):
-                if not self._cancel and (s == cups.HTTP_UNAUTHORIZED or
-                                         s == cups.HTTP_FORBIDDEN):
+                if not self._cancel:
                     self._failed (s == cups.HTTP_FORBIDDEN)
                 else:
                     raise
@@ -445,6 +473,10 @@ class Connection:
     def _on_authentication_response (self, dialog, response):
         (self._use_user,
          self._use_password) = dialog.get_auth_info ()
+        global_authinfocache.cache_auth_info ((self._use_user,
+                                               self._use_password),
+                                              host=self._server,
+                                              port=self._port)
         dialog.destroy ()
 
         if (response == gtk.RESPONSE_CANCEL or
