@@ -739,6 +739,18 @@ device_id_from_bluetooth (const char *bdaddr, struct device_id *id)
   g_free (device_id);
 }
 
+static char *
+uri_from_bdaddr (const char *devpath)
+{
+  return g_strdup_printf("bluetooth://%c%c%c%c%c%c%c%c%c%c%c%c",
+			 devpath[0], devpath[1],
+			 devpath[3], devpath[4],
+			 devpath[6], devpath[7],
+			 devpath[9], devpath[10],
+			 devpath[12], devpath[13],
+			 devpath[15], devpath[16]);
+}
+
 static const char *
 no_password (const char *prompt)
 {
@@ -1366,6 +1378,7 @@ do_add (const char *cmd, const char *devpath)
   char *usb_device_devpath = NULL;
   char usbserial[256];
   char usblpdev[8] = "";
+  gboolean is_bluetooth;
 
   if (getenv ("DEBUG") == NULL)
     {
@@ -1391,8 +1404,10 @@ do_add (const char *cmd, const char *devpath)
 
   syslog (LOG_DEBUG, "add %s", devpath);
 
+  is_bluetooth = bluetooth_verify_address (devpath);
+
   map = read_usb_uri_map ();
-  if (bluetooth_verify_address (devpath)) {
+  if (is_bluetooth) {
     usbserial[0] = '\0';
     device_id_from_bluetooth (devpath, &id);
   } else {
@@ -1403,7 +1418,7 @@ do_add (const char *cmd, const char *devpath)
 
   if (!id.mfg || !id.mdl)
     {
-      syslog (LOG_ERR, "invalid or missing IEEE 1284 Device ID%s%s",
+      syslog (LOG_DEBUG, "invalid or missing IEEE 1284 Device ID%s%s",
 	      id.full_device_id ? " " : "",
 	      id.full_device_id ? id.full_device_id : "");
       exit (1);
@@ -1412,11 +1427,22 @@ do_add (const char *cmd, const char *devpath)
   syslog (LOG_DEBUG, "MFG:%s MDL:%s SERN:%s serial:%s", id.mfg, id.mdl,
 	  id.sern ? id.sern : "-", usbserial[0] ? usbserial : "-");
 
-  find_matching_device_uris (&id, usbserial, &device_uris, usb_device_devpath,
-			     map);
-  free (usb_device_devpath);
+  if (!is_bluetooth)
+    {
+      find_matching_device_uris (&id, usbserial, &device_uris, usb_device_devpath,
+				 map);
+      free (usb_device_devpath);
+    } else {
+      char *device_uri;
+
+      device_uri = uri_from_bdaddr (devpath);
+      add_device_uri (&device_uris, device_uri);
+      g_free (device_uri);
+    }
+
   if (device_uris.n_uris == 0)
     {
+      syslog (LOG_ERR, "no corresponding CUPS device found");
       free_device_id (&id);
       return 0;
     }
@@ -1471,14 +1497,44 @@ do_add (const char *cmd, const char *devpath)
 }
 
 static void
-disable_queue (const char *printer_uri, void *context)
+remove_queue (const char *printer_uri)
 {
   /* Disable it. */
   http_t *cups = httpConnectEncrypt (cupsServer (), ippPort (),
 				     cupsEncryption ());
   ipp_t *request, *answer;
 
-  //FIXME remove instead of disable for Bluetooth
+  if (cups == NULL)
+    return;
+
+  request = ippNewRequest (CUPS_DELETE_PRINTER);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
+		"printer-uri", NULL, printer_uri);
+  answer = cupsDoRequest (cups, request, "/admin/");
+  if (!answer)
+    {
+      syslog (LOG_ERR, "Failed to send IPP-Delete-Printer request");
+      httpClose (cups);
+      return;
+    }
+
+  if (answer->request.status.status_code > IPP_OK_CONFLICT)
+    syslog (LOG_ERR, "IPP-Delete-Printer request failed");
+  else
+    syslog (LOG_INFO, "Deleted printer %s as the corresponding device "
+	    "was unpaired", printer_uri);
+
+  ippDelete (answer);
+  httpClose (cups);
+}
+
+static void
+disable_queue (const char *printer_uri, void *context)
+{
+  /* Disable it. */
+  http_t *cups = httpConnectEncrypt (cupsServer (), ippPort (),
+				     cupsEncryption ());
+  ipp_t *request, *answer;
 
   if (cups == NULL)
     return;
@@ -1515,6 +1571,16 @@ do_remove (const char *devpath)
   char usblpdev[8] = "";
   syslog (LOG_DEBUG, "remove %s", devpath);
 
+  if (bluetooth_verify_address (devpath))
+    {
+      char *device_uri;
+
+      device_uri = uri_from_bdaddr (devpath);
+      remove_queue (devpath);
+      g_free (device_uri);
+      return 0;
+    }
+
   map = read_usb_uri_map ();
   prev = &map->entries;
   for (entry = map->entries; entry; entry = entry->next)
